| Safe Haskell | None |
|---|---|
| Language | GHC2021 |
Polysemy.Resume.Effect.Stop
Description
Documentation
data Stop e (a :: Type -> Type) b where Source #
An effect similar to Error without the ability to be caught.
Used to signal that an error is supposed to be expected by dependent programs.
interpretStopper ::
Member (Stop Boom) r =>
InterpreterFor Stopper r
interpretStopper =
interpret \case
StopBang -> stop (Bang 13)
StopBoom -> stop (Boom "ouch")